﻿{ **************************************************************************** }
{                                                                              }
{                      汎用ビットフィールド操作レコード                        }
{              Copyright (c) 2012-2015 Hideaki Tominaga (DEKO)                 }
{                                                                              }
{ ---------------------------------------------------------------------------- }
{                         http://ht-deko.minim.ne.jp/                          }
{                           deko@ht-deko.minim.ne.jp                           }
{ ---------------------------------------------------------------------------- }
{                                                                              }
{ USAGE:                                                                       }
{   http://ht-deko.minim.ne.jp/tech076.html                                    }
{                                                                              }
{ **************************************************************************** }
unit uBitField;

interface

uses
  SysUtils, Types;

type
  { TBit32 }
  TBit32 = 0..31;

  { TByte32 }
  TByte32 = 0..3;

  { TCnt32 }
  TCnt32 = 1..32;

  { TBits32 }
  TBits32 = set of TBit32;

  { PBitFieldRec }
  PBitFieldRec = ^TBitFieldRec;

  { TBitFieldRec }
  TBitFieldRec = packed record
    // 型変換 (代入時)
    class operator Implicit(a: DWORD): TBitFieldRec;                            // 暗黙の型変換: TBitFieldRec に DWORD を代入 (レコードが左辺)
    class operator Implicit(a: TBitFieldRec): DWORD;                            // 暗黙の型変換: DWORD に TBitFieldRec を代入 (レコードが右辺)
    // 算術演算子 (単項)
    class operator Negative(a: TBitFieldRec): TBitFieldRec;                     // 符号反転: -
    class operator Positive(a: TBitFieldRec): TBitFieldRec;                     // 符号恒等: +
    // 算術演算子
    class operator Add(a, b: TBitFieldRec): TBitFieldRec;                       // 加算: +
    class operator Subtract(a, b: TBitFieldRec): TBitFieldRec;                  // 減算: -
    class operator Multiply(a, b: TBitFieldRec): TBitFieldRec;                  // 乗算: *
    class operator IntDivide(a, b: TBitFieldRec): TBitFieldRec;                 // 整数除算: div
    class operator Modulus(a, b: TBitFieldRec): TBitFieldRec;                   // 剰余: mod
    // 論理演算子
    class operator LogicalNot(a: TBitFieldRec): TBitFieldRec;                                 // 否定: not
    class operator LogicalAnd(a, b: TBitFieldRec): Boolean;                     // 論理積: and
    class operator LogicalOr(a, b: TBitFieldRec): Boolean;                      // 論理和: or
    class operator LogicalXor(a, b: TBitFieldRec): Boolean;                     // 排他的論理和: xor
    // 論理（ビット）演算子
    class operator BitwiseAnd(a, b: TBitFieldRec): TBitFieldRec;                // ビット and: and
    class operator BitwiseOr(a, b: TBitFieldRec): TBitFieldRec;                 // ビット or: or
    class operator BitwiseXor(a, b: TBitFieldRec): TBitFieldRec;                // ビット xor: xor
    class operator LeftShift(a, b: TBitFieldRec): TBitFieldRec;                 // ビット単位の左シフト: shl
    class operator RightShift(a, b: TBitFieldRec): TBitFieldRec;                // ビット単位の右シフト: shr
    // 関係演算子
    class operator Equal(a, b: TBitFieldRec): Boolean;                          // 等しい: =
    class operator NotEqual(a, b: TBitFieldRec): Boolean;                       // 等しくない: <>
    class operator LessThan(a, b: TBitFieldRec): Boolean;                       // より小さい: <
    class operator GreaterThan(a, b: TBitFieldRec): Boolean;                    // より大きい: >
    class operator LessThanOrEqual(a, b: TBitFieldRec): Boolean;                // 以下: <=
    class operator GreaterThanOrEqual(a, b: TBitFieldRec): Boolean;             // 以上: >=
    // インクリメント / デクリメント
    class operator Inc(a: TBitFieldRec): TBitFieldRec;                          // インクリメント: Inc()
    class operator Dec(a: TBitFieldRec): TBitFieldRec;                          // デクリメント: Dec()
    {$IFDEF CONDITIONALEXPRESSIONS}
    {$IF CompilerVersion >= 20.0}
    // Include / Exclude
    class operator Include(a: TBitFieldRec; b: TBits32): TBitFieldRec;          // 含める: Include()  (2009 or later)
    class operator Exclude(a: TBitFieldRec; b: TBits32): TBitFieldRec;          // 含めない: Exclude()  (2009 or later)
    // 集合演算子
    class operator In(a: TBit32; b: TBitFieldRec): Boolean;                     // メンバかどうか: in  (2009 or later)
    {$IFEND}
    {$ENDIF}
  private
    function GetBit(Index: TBit32): Boolean;
    procedure SetBit(Index: TBit32; const aValue: Boolean);
  public
    function GetBitField(aStartBit: TBit32; aCount: TCnt32): DWORD;             // ビットフィールドの値を取得
    procedure SetBitField(aStartBit: TBit32; aCount: TCnt32; aValue: DWORD);    // ビットフィールドへ値を設定
    function ToString: string;                                                  // 10 進数文字列へ変換
    function ToHexString(Digits: Integer = 1): string;                          // 16 進数文字列へ変換
    function ToOctString(Digits: Integer = 1): string;                          // 8 進数文字列へ変換
    function ToBinString(Digits: Integer = 1): string;                          // 2 進数文字列へ変換
    property Bits[Index: TBit32]: Boolean read GetBit write SetBit; default;    // ブール型ビット値 (bit0..bit31)
    procedure SwapByte(A, B: TByte32);                                          // 任意のバイトを入れ替える
    procedure SwapWord;                                                         // 上位ワードと下位ワードを入れ替える
  case BYTE of                                                                  // ペイロード共用体
    0: (Value: DWORD);                                                          //  - DWORD
    1: (Words: array [0..1] of WORD);                                           //  - ワード配列
    2: (LoWord, HiWord: WORD);                                                  //  - Lo / Hi ワード
    3: (Bytes: array [0..3] of BYTE);                                           //  - バイト配列
  end;

const
  bit00 =  0; bit0 = 0;
  bit01 =  1; bit1 = 1;
  bit02 =  2; bit2 = 2;
  bit03 =  3; bit3 = 3;
  bit04 =  4; bit4 = 4;
  bit05 =  5; bit5 = 5;
  bit06 =  6; bit6 = 6;
  bit07 =  7; bit7 = 7;
  bit08 =  8; bit8 = 8;
  bit09 =  9; bit9 = 9;
  bit10 = 10;
  bit11 = 11;
  bit12 = 12;
  bit13 = 13;
  bit14 = 14;
  bit15 = 15;
  bit16 = 16;
  bit17 = 17;
  bit18 = 18;
  bit19 = 19;
  bit20 = 20;
  bit21 = 21;
  bit22 = 22;
  bit23 = 23;
  bit24 = 24;
  bit25 = 25;
  bit26 = 26;
  bit27 = 27;
  bit28 = 28;
  bit29 = 29;
  bit30 = 30;
  bit31 = 31;

  BYTE0 =  0;
  BYTE1 =  1;
  BYTE2 =  2;
  BYTE3 =  3;
  
implementation

{ TBitFieldRec }

// 型変換 (代入時)
class operator TBitFieldRec.Implicit(a: DWORD): TBitFieldRec;
begin
  result.Value := a;
end;

class operator TBitFieldRec.Implicit(a: TBitFieldRec): DWORD;
begin
  result := a.Value;
end;

// 算術演算子 (単項)
class operator TBitFieldRec.Negative(a: TBitFieldRec): TBitFieldRec;
begin
  result.Value := -a.Value;
end;

class operator TBitFieldRec.Positive(a: TBitFieldRec): TBitFieldRec;
begin
  result.Value := +a.Value;
end;

// 算術演算子
class operator TBitFieldRec.Add(a, b: TBitFieldRec): TBitFieldRec;
begin
  result.Value := a.Value + b.Value;
end;

class operator TBitFieldRec.Subtract(a, b: TBitFieldRec): TBitFieldRec;
begin
  result.Value := a.Value - b.Value;
end;

class operator TBitFieldRec.Multiply(a, b: TBitFieldRec): TBitFieldRec;
begin
  result.Value := a.Value * b.Value;
end;

class operator TBitFieldRec.IntDivide(a, b: TBitFieldRec): TBitFieldRec;
begin
  result.Value := a.Value div b.Value;
end;

class operator TBitFieldRec.Modulus(a, b: TBitFieldRec): TBitFieldRec;
begin
  result.Value := a.Value mod b.Value;
end;

// 論理演算子
class operator TBitFieldRec.LogicalNot(a: TBitFieldRec): TBitFieldRec;
begin
  result.Value := not a.Value;
end;

class operator TBitFieldRec.LogicalAnd(a, b: TBitFieldRec): Boolean;
begin
  result := (a.Value and b.Value) > 0;
end;

class operator TBitFieldRec.LogicalOr(a, b: TBitFieldRec): Boolean;
begin
  result := (a.Value or b.Value) > 0;
end;

class operator TBitFieldRec.LogicalXor(a, b: TBitFieldRec): Boolean;
begin
  result := (a.Value xor b.Value) > 0;
end;

// 論理（ビット）演算子
class operator TBitFieldRec.BitwiseAnd(a, b: TBitFieldRec): TBitFieldRec;
begin
  result.Value := (a.Value and b.Value);
end;

class operator TBitFieldRec.BitwiseOr(a, b: TBitFieldRec): TBitFieldRec;
begin
  result.Value := (a.Value or b.Value);
end;

class operator TBitFieldRec.BitwiseXor(a, b: TBitFieldRec): TBitFieldRec;
begin
  result.Value := (a.Value xor b.Value);
end;

class operator TBitFieldRec.LeftShift(a, b: TBitFieldRec): TBitFieldRec;
begin
  result.Value := a.Value shl b.Value;
end;

class operator TBitFieldRec.RightShift(a, b: TBitFieldRec): TBitFieldRec;
begin
  result.Value := a.Value shr b.Value;
end;

// 関係演算子
class operator TBitFieldRec.Equal(a, b: TBitFieldRec): Boolean;
begin
  result := (a.Value = b.Value);
end;

class operator TBitFieldRec.NotEqual(a, b: TBitFieldRec): Boolean;
begin
  result := (a.Value <> b.Value);
end;

class operator TBitFieldRec.GreaterThan(a, b: TBitFieldRec): Boolean;
begin
  result := (a.Value > b.Value);
end;

class operator TBitFieldRec.GreaterThanOrEqual(a, b: TBitFieldRec): Boolean;
begin
  result := (a.Value >= b.Value);
end;

class operator TBitFieldRec.LessThan(a, b: TBitFieldRec): Boolean;
begin
  result := (a.Value < b.Value);
end;

class operator TBitFieldRec.LessThanOrEqual(a, b: TBitFieldRec): Boolean;
begin
  result := (a.Value <= b.Value);
end;

// インクリメント / デクリメント
class operator TBitFieldRec.Inc(a: TBitFieldRec): TBitFieldRec;
begin
  result.Value := a.Value + 1;
end;

class operator TBitFieldRec.Dec(a: TBitFieldRec): TBitFieldRec;
begin
  result.Value := a.Value - 1;
end;

{$IFDEF CONDITIONALEXPRESSIONS}
{$IF CompilerVersion >= 20.0}

// Include / Exclude (2009 or later)
class operator TBitFieldRec.Include(a: TBitFieldRec; b: TBits32): TBitFieldRec;
begin
  result.Value := a.Value or DWORD(b);
end;

class operator TBitFieldRec.Exclude(a: TBitFieldRec; b: TBits32): TBitFieldRec;
begin
  result.Value := a.Value and (not DWORD(b));
end;

// 集合演算子 (2009 or later)
class operator TBitFieldRec.In(a: TBit32; b: TBitFieldRec): Boolean;
begin
  result := a in TBits32(b.Value);
end;
{$IFEND}
{$ENDIF}

// Bits プロパティ
function TBitFieldRec.GetBit(Index: TBit32): Boolean;
begin
  result := ((Value and (1 shl Index)) > 0);
end;

procedure TBitFieldRec.SetBit(Index: TBit32; const aValue: Boolean);
begin
  if aValue then
    Value := Value or (1 shl Index)
  else
    Value := Value and (not (1 shl Index));
end;

// 最下位ビットを返す
function GetEndBit(aStartBit: TBit32; aCount: TCnt32): DWORD;
var
  EndBit: Integer;
begin
  EndBit := aStartBit - aCount + 1;
  if EndBit < 0 then
    EndBit := 0;
  result := EndBit;
end;

// ビットマスクを作成
function CreateBitMask(aStartBit: TBit32; aCount: TCnt32): DWORD;
var
  i: Integer;
  EndBit: DWORD;
begin
  result := 0;
  EndBit := GetEndBit(aStartBit, aCount);
  for i:= aStartBit downto EndBit do
    Inc(result, 1 shl i);
end;

// ビットフィールドの値を取得
function TBitFieldRec.GetBitField(aStartBit: TBit32; aCount: TCnt32): DWORD;
var
  BitMask: DWORD;
  EndBit: DWORD;
begin
  BitMask := CreateBitMask(aStartBit, aCount);
  EndBit := GetEndBit(aStartBit, aCount);
  result := (Value and BitMask) shr EndBit;
end;

// ビットフィールドへ値を設定
procedure TBitFieldRec.SetBitField(aStartBit: TBit32; aCount: TCnt32;
  aValue: DWORD);
var
  EndBit: DWORD;
  BitMask: DWORD;
begin
  BitMask := not CreateBitMask(aStartBit, aCount);
  EndBit := GetEndBit(aStartBit, aCount);
  Value := (Value and BitMask) or (aValue shl EndBit);
end;

// 10 進数文字列へ変換
function TBitFieldRec.ToString: string;
begin
  result := IntToStr(Value);
end;

// 16 進数文字列へ変換
function TBitFieldRec.ToHexString(Digits: Integer): string;
begin
  result := IntToHex(Value, Digits);
end;

// 8 進数文字列へ変換
function TBitFieldRec.ToOctString(Digits: Integer): string;
var
  i: Integer;
  dValue: UInt32;
begin
  dValue := Self.Value;
  for i:=0 to 10 do
    begin
      result := IntToStr(dValue and $00000007) + result;
      dValue := dValue shr 3;
      if (dValue = 0) and (i >= Digits) then
        Break;
    end;
end;

// 2 進数文字列へ変換
function TBitFieldRec.ToBinString(Digits: Integer): string;
var
  i: Integer;
  dValue: DWORD;
begin
  dValue := Value;
  result := '';
  for i:=1 to 32 do
    begin
      result := IntToStr(dValue and 1) + result;
      dValue := dValue shr 1;
      if (dValue = 0) and (i >= Digits) then
        Break;
    end;
end;

// 任意のバイトを入れ替える
procedure TBitFieldRec.SwapByte(A, B: TByte32);
var
  Swap: Byte;
begin
  if a = B then
    Exit;
  Swap := Self.Bytes[A];
  Self.Bytes[A] := Self.Bytes[B];
  Self.Bytes[B] := Swap;
end;

// 上位ワードと下位ワードを入れ替える
procedure TBitFieldRec.SwapWord;
var
  Swap: Word;
begin
  Swap := Self.LoWord;
  Self.LoWord := Self.HiWord;
  Self.HiWord := Swap;
end;

end.
